from ninja.security.base import AuthBase
from ninja.compatibility.request import get_headers


class APIKeyBase(AuthBase):
    openapi_type = "apiKey"
    param_name = "key"

    def __init__(self):
        self.openapi_name = self.param_name
        super().__init__()

    def __call__(self, request):
        key = self._get_key(request)
        return self.authenticate(request, key)

    def authenticate(self, request, key):
        raise NotImplementedError("Please implement authenticate(self, request, key)")


class APIKeyQuery(APIKeyBase):
    openapi_in = "query"

    def _get_key(self, request):
        return request.GET.get(self.param_name)


class APIKeyCookie(APIKeyBase):
    openapi_in = "cookie"

    def _get_key(self, request):
        return request.COOKIES.get(self.param_name)


class APIKeyHeader(APIKeyBase):
    openapi_in = "header"

    def _get_key(self, request):
        headers = get_headers(request)
        return headers.get(self.param_name)



from datetime import date
from typing import Union
from client import NinjaClient
from ninja import Router

router = Router()


@router.get("/test")
def view(request, value: Union[date, str]):
    return [value, type(value).__name__]


client = NinjaClient(router)


def test_union():
    assert client.get("/test?value=today").json() == ["today", "str"]
    assert client.get("/test?value=2020-01-15").json() == ["2020-01-15", "date"]
    # TODO: test also schema



import json
from pydantic import BaseModel
from django.http import QueryDict
from ninja.compatibility import get_headers
from ninja.errors import InvalidBodyJson


class ParamModel(BaseModel):
    @classmethod
    def resolve(cls, request, path_params):
        data = cls.get_request_data(request, path_params)
        if data is None:
            return cls()

        varname = getattr(cls, "_single_attr", None)
        if varname:
            data = {varname: data}
        # TODO: I guess if data is not dict - raise an InvalidInput
        return cls(**data)


class QueryModel(ParamModel):
    @classmethod
    def get_request_data(cls, request, path_params):
        return _querydict_to_dict(cls, request.GET)


class PathModel(ParamModel):
    @classmethod
    def get_request_data(cls, request, path_params):
        return path_params


class HeaderModel(ParamModel):
    @classmethod
    def get_request_data(cls, request, path_params):
        data = {}
        headers = get_headers(request)
        for name, field in cls.__fields__.items():
            if name in headers:
                data[name] = headers[name]
            elif field.alias in headers:
                data[field.alias] = headers[field.alias]
        return data


class CookieModel(ParamModel):
    @classmethod
    def get_request_data(cls, request, path_params):
        return request.COOKIES


class BodyModel(ParamModel):
    @classmethod
    def get_request_data(cls, request, path_params):
        if request.body:
            try:
                # TODO: maybe better to cache data on request instance if multiple bodies parsing same large body
                return json.loads(request.body)
            except Exception as e:
                raise InvalidBodyJson(e)


class FormModel(ParamModel):
    @classmethod
    def get_request_data(cls, request, path_params):
        return _querydict_to_dict(cls, request.POST)


def _querydict_to_dict(cls, data: QueryDict):
    list_fields = getattr(cls, "_collection_fields", [])
    result = {}
    for key in data.keys():
        if key in list_fields:
            result[key] = data.getlist(key)
        else:
            result[key] = data[key]
    return result



from unittest.mock import patch
from ninja import NinjaAPI
from client import NinjaClient


def test_examples():

    api = NinjaAPI()

    with patch("builtins.api", api, create=True):
        import docs.src.tutorial.path.code010

        client = NinjaClient(api)

        response = client.get("/events/2020/1/1")
        assert response.json() == {"date": "2020-01-01"}
        schema = api.get_openapi_schema("")
        events_params = schema["paths"]["/events/{year}/{month}/{day}"]["get"][
            "parameters"
        ]
        assert events_params == [
            {"in": "path", "name": "year", "required": True},
            {"in": "path", "name": "month", "required": True},
            {"in": "path", "name": "day", "required": True},
        ]



import datetime
from ninja import Schema, Query


class Filters(Schema):
    limit: int = 100
    offset: int = None
    query: str = None


@api.get("/filter")
def events(request, filters: Filters = Query(...)):
    return {"filters": filters.dict()}



from collections import OrderedDict
from typing import Callable, List

from django.http import HttpResponseNotAllowed
from django.urls import path as django_path
from ninja.operation import Operation, PathView
from ninja.constants import NOT_SET


class Router:
    def __init__(self):
        self.operations = OrderedDict()

    def get(self, path: str, *, auth=NOT_SET):
        return self.api_operation(["GET"], path, auth=auth)

    def post(self, path: str, *, auth=NOT_SET):
        return self.api_operation(["POST"], path, auth=auth)

    def delete(self, path: str, *, auth=NOT_SET):
        return self.api_operation(["DELETE"], path, auth=auth)

    def patch(self, path: str, *, auth=NOT_SET):
        return self.api_operation(["PATCH"], path, auth=auth)

    def put(self, path: str, *, auth=NOT_SET):
        return self.api_operation(["PUT"], path, auth=auth)

    def api_operation(self, methods: List[str], path: str, *, auth=NOT_SET):
        def decorator(view_func):
            self.add_api_operation(path, methods, view_func, auth=auth)
            return view_func

        return decorator

    def add_api_operation(
        self, path: str, methods: List[str], view_func: Callable, *, auth=NOT_SET
    ):
        if path not in self.operations:
            self.operations[path] = PathView()
        self.operations[path].append(
            path=path, methods=methods, view_func=view_func, auth=auth
        )

    def urls_paths(self, prefix: str):
        for path, path_view in self.operations.items():
            path = path.replace("{", "<").replace("}", ">")
            route = "/".join([i for i in (prefix, path) if i])
            # to skip lot of checks we simply treat doulbe slash as a mistake:
            route = route.replace("//", "/")
            route = route.lstrip("/")

            yield django_path(route, path_view.get_view())



from pydantic import BaseModel
from ninja.signature.details import is_pydantic_model


def test_is_pydantic_model():
    class Model(BaseModel):
        x: int

    assert is_pydantic_model(Model)
    assert is_pydantic_model("instance") is False



import pytest
from unittest.mock import Mock
from ninja import NinjaAPI
from ninja.security import (
    django_auth,
    APIKeyCookie,
    APIKeyHeader,
    APIKeyQuery,
    HttpBearer,
    HttpBasicAuth,
)
from ninja.security.base import AuthBase
from ninja.errors import ConfigError
from client import NinjaClient


def callble_auth(request):
    return request.GET.get("auth")


class KeyQuery(APIKeyQuery):
    def authenticate(self, request, key):
        if key == "keyquerysecret":
            return key


class KeyHeader(APIKeyHeader):
    def authenticate(self, request, key):
        if key == "keyheadersecret":
            return key


class KeyCookie(APIKeyCookie):
    def authenticate(self, request, key):
        if key == "keycookiersecret":
            return key


class BasicAuth(HttpBasicAuth):
    def authenticate(self, request, username, password):
        if username == "admin" and password == "secret":
            return username


class BearerAuth(HttpBearer):
    def authenticate(self, request, token):
        if token == "bearertoken":
            return token


def demo_operaiton(request):
    return {"auth": request.auth}


api = NinjaAPI()

for path, auth in [
    ("django_auth", django_auth),
    ("callable", callble_auth),
    ("apikeyquery", KeyQuery()),
    ("apikeyheader", KeyHeader()),
    ("apikeycookie", KeyCookie()),
    ("basic", BasicAuth()),
    ("bearer", BearerAuth()),
]:
    api.get(f"/{path}", auth=auth)(demo_operaiton)


client = NinjaClient(api)


class MockUser(str):
    is_authenticated = True


@pytest.mark.parametrize(
    "path,kwargs,expected_code",
    [
        ("/django_auth", {}, 401),
        ("/django_auth", dict(user=MockUser("admin")), 200),
        ("/callable", {}, 401),
        ("/callable?auth=demo", {}, 200),
        ("/apikeyquery", {}, 401),
        ("/apikeyquery?key=keyquerysecret", {}, 200),
        ("/apikeyheader", {}, 401),
        ("/apikeyheader", dict(headers={"key": "keyheadersecret"}), 200),
        ("/apikeycookie", {}, 401),
        ("/apikeycookie", dict(COOKIES={"key": "keycookiersecret"}), 200),
        ("/basic", {}, 401),
        ("/basic", dict(headers={"Authorization": "Basic YWRtaW46c2VjcmV0"}), 200),
        ("/basic", dict(headers={"Authorization": "YWRtaW46c2VjcmV0"}), 200),
        ("/basic", dict(headers={"Authorization": "Basic invalid"}), 401),
        ("/basic", dict(headers={"Authorization": "some invalid value"}), 401),
        ("/bearer", {}, 401),
        ("/bearer", dict(headers={"Authorization": "Bearer bearertoken"}), 200),
        ("/bearer", dict(headers={"Authorization": "Invalid bearertoken"}), 401),
    ],
)
def test_auth(path, kwargs, expected_code):
    response = client.get(path, **kwargs)
    assert response.status_code == expected_code


def test_schema():
    schema = api.get_openapi_schema()
    assert schema["components"]["securitySchemes"] == {
        "BasicAuth": {"scheme": "basic", "type": "http"},
        "BearerAuth": {"scheme": "bearer", "type": "http"},
        "KeyCookie": {"in": "cookie", "name": "key", "type": "apiKey"},
        "KeyHeader": {"in": "header", "name": "key", "type": "apiKey"},
        "KeyQuery": {"in": "query", "name": "key", "type": "apiKey"},
    }
    # TODO: Samename for schema check
    # TOOD: check operation security attributes


def test_invalid_setup():
    request = Mock()
    headers = {"Authorization": "Bearer test"}
    request.META = {"HTTP_" + k: v for k, v in headers.items()}
    request.headers = headers

    class MyAuth1(AuthBase):
        pass

    class MyAuth2(AuthBase):
        openapi_type = "my"

    with pytest.raises(ConfigError):
        MyAuth1()(request)
    with pytest.raises(NotImplementedError):
        MyAuth2()(request)
    with pytest.raises(NotImplementedError):
        APIKeyCookie()(request)
    with pytest.raises(NotImplementedError):
        APIKeyHeader()(request)
    with pytest.raises(NotImplementedError):
        APIKeyQuery()(request)
    with pytest.raises(NotImplementedError):
        HttpBearer()(request)

    headers = {"Authorization": "Basic YWRtaW46c2VjcmV0"}
    request.META = {"HTTP_" + k: v for k, v in headers.items()}
    request.headers = headers

    with pytest.raises(NotImplementedError):
        HttpBasicAuth()(request)



from pydantic.fields import FieldInfo, ModelField
from pydantic import BaseConfig
from typing import Any
from ninja import params_models


class Param(FieldInfo):
    def __init__(
        self,
        default: Any,
        *,
        alias: str = None,
        title: str = None,
        description: str = None,
        gt: float = None,
        ge: float = None,
        lt: float = None,
        le: float = None,
        min_length: int = None,
        max_length: int = None,
        regex: str = None,
        deprecated: bool = None,
        # param_name: str = None,
        # param_type: Any = None,
        **extra: Any,
    ):
        # print('alias = ', alias)
        self.deprecated = deprecated
        # self.param_name: str = None
        # self.param_type: Any = None
        self.model_field: ModelField = None
        super().__init__(
            default,
            alias=alias,
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            **extra,
        )

    @classmethod
    def _in(cls):
        "Openapi param.in value"
        return cls.__name__.lower()


class Path(Param):
    _model = params_models.PathModel


class Query(Param):
    _model = params_models.QueryModel


class Header(Param):
    _model = params_models.HeaderModel


class Cookie(Param):
    _model = params_models.CookieModel


class Body(Param):
    _model = params_models.BodyModel


class Form(Param):
    _model = params_models.FormModel



import re
import inspect
import asyncio
from typing import Callable, Dict, Any, Set
from pydantic.typing import ForwardRef, evaluate_forwardref


def get_typed_signature(call: Callable) -> inspect.Signature:
    "Finds call signature and resolves all forwardrefs"
    signature = inspect.signature(call)
    globalns = getattr(call, "__globals__", {})
    typed_params = [
        inspect.Parameter(
            name=param.name,
            kind=param.kind,
            default=param.default,
            annotation=get_typed_annotation(param, globalns),
        )
        for param in signature.parameters.values()
    ]
    return_annotation = signature.return_annotation
    if isinstance(return_annotation, str):
        return_annotation = make_forwardref(return_annotation, globalns)

    typed_signature = inspect.Signature(
        typed_params, return_annotation=return_annotation
    )
    return typed_signature


def get_typed_annotation(param: inspect.Parameter, globalns: Dict[str, Any]) -> Any:
    annotation = param.annotation
    if isinstance(annotation, str):
        annotation = make_forwardref(annotation, globalns)
    return annotation


def make_forwardref(annotation: str, globalns: Dict[str, Any]):
    annotation = ForwardRef(annotation)
    annotation = evaluate_forwardref(annotation, globalns, globalns)
    return annotation


def get_path_param_names(path: str) -> Set[str]:
    "turns path string like /foo/{var}/path/{another}/end to set ['var', 'another']"
    return {item.strip("{}") for item in re.findall("{[^}]*}", path)}


def is_async(callable: Callable):
    return asyncio.iscoroutinefunction(callable)



from ninja import NinjaAPI
from ninja.security import APIKeyQuery
from client import NinjaClient


class KeyQuery1(APIKeyQuery):
    def authenticate(self, request, key):
        if key == "k1":
            return key


class KeyQuery2(APIKeyQuery):
    def authenticate(self, request, key):
        if key == "k2":
            return key


api = NinjaAPI(auth=KeyQuery1())


@api.get("/default")
def default(request):
    return {"auth": request.auth}


@api.api_operation(["POST", "PATCH"], "/multi-method-no-auth")
def multi_no_auth(request):
    return {"auth": request.auth}


@api.api_operation(["POST", "PATCH"], "/multi-method-auth", auth=KeyQuery2())
def multi_auth(request):
    return {"auth": request.auth}


client = NinjaClient(api)


def test_multi():
    assert client.get("/default").status_code == 401
    assert client.get("/default?key=k1").json() == {"auth": "k1"}

    assert client.post("/multi-method-no-auth").status_code == 401
    assert client.post("/multi-method-no-auth?key=k1").json() == {"auth": "k1"}

    assert client.patch("/multi-method-no-auth").status_code == 401
    assert client.patch("/multi-method-no-auth?key=k1").json() == {"auth": "k1"}

    assert client.post("/multi-method-auth?key=k1").status_code == 401
    assert client.patch("/multi-method-auth?key=k1").status_code == 401

    assert client.post("/multi-method-auth?key=k2").json() == {"auth": "k2"}
    assert client.patch("/multi-method-auth?key=k2").json() == {"auth": "k2"}



from datetime import date
from enum import Enum
from pydantic import BaseModel
from ninja import Router
from client import NinjaClient


class RoomEnum(str, Enum):
    double = "double"
    twin = "twin"
    single = "single"


class Booking(BaseModel):
    start: date
    end: date
    room: RoomEnum = RoomEnum.double


router = Router()


@router.post("/book")
def booking_request(request, booking: Booking):
    return booking


@router.get("/search")
def booking_search(request, date: date, room: RoomEnum):
    return {"date": date, "room": room}


client = NinjaClient(router)


def test_enums():
    response = client.post(
        "/book", json={"start": "2020-01-01", "end": "2020-01-02", "room": "double"}
    )
    assert response.status_code == 200, response.content
    assert response.json() == {
        "start": "2020-01-01",
        "end": "2020-01-02",
        "room": "double",
    }

    response = client.post(
        "/book", json={"start": "2020-01-01", "end": "2020-01-02", "room": "triple"}
    )
    assert response.status_code == 422

    response = client.get("/search?date=2020-01-01&room=twin")
    assert response.status_code == 200
    assert response.json() == {
        "date": "2020-01-01",
        "room": "twin",
    }

    response = client.get("/search?date=2020-01-01&room=other")
    assert response.status_code == 422



from base64 import b64decode
from urllib.parse import unquote
from ninja.compatibility import get_headers
from ninja.security.base import AuthBase


class HttpAuthBase(AuthBase):
    openapi_type = "http"


class HttpBearer(HttpAuthBase):
    openapi_scheme = "bearer"
    header = "Authorization"

    def __call__(self, request):
        headers = get_headers(request)
        auth_value = headers.get(self.header)
        if not auth_value:
            return
        parts = auth_value.split(" ")

        if parts[0].lower() != "bearer":
            print(f"Unexpected auth - '{auth_value}'")
            return
        token = " ".join(parts[1:])
        return self.authenticate(request, token)

    def authenticate(self, request, token):
        raise NotImplementedError("Please implement authenticate(self, request, token)")


class DecodeError(Exception):
    pass


class HttpBasicAuth(HttpAuthBase):  # TODO: maybe HttpBasicAuthBase
    openapi_scheme = "basic"
    header = "Authorization"

    def __call__(self, request):
        headers = get_headers(request)
        auth_value = headers.get(self.header)
        if not auth_value:
            return

        try:
            username, password = self.decode_authorization(auth_value)
        except DecodeError as e:
            print(e)
            return
        return self.authenticate(request, username, password)

    def authenticate(self, request, username, password):
        raise NotImplementedError(
            "Please implement authenticate(self, request, username, password)"
        )

    def decode_authorization(self, value):
        parts = value.split(" ")
        if len(parts) == 1:
            user_pass_encoded = parts[0]
        elif len(parts) == 2 and parts[0].lower() == "basic":
            user_pass_encoded = parts[1]
        else:
            raise DecodeError("Invlid Authorization header")

        try:
            username, password = b64decode(user_pass_encoded).decode().split(":", 1)
            return unquote(username), unquote(password)
        except Exception as e:
            print(e)
            raise DecodeError("Invlid Authorization header") from e



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['ninja',
 'ninja.compatibility',
 'ninja.openapi',
 'ninja.security',
 'ninja.signature']

package_data = \
{'': ['*'],
 'ninja': ['.pytest_cache/*',
           '.pytest_cache/v/*',
           '.pytest_cache/v/cache/*',
           'templates/*',
           'templates/ninja/*']}

install_requires = \
['Django >=2.0.13', 'pydantic >=1.4,<2.0.0']

extras_require = \
{'doc': ['mkdocs', 'mkdocs-material', 'markdown-include'],
 'test': ['pytest', 'pytest-cov', 'pytest-django', 'pytest-asyncio', 'black']}

setup(name='django-ninja',
      version='0.2.0',
      description='Django Ninja - Fast Django REST framework',
      author='Vitaliy Kucheryaviy',
      author_email='ppr.vitaly@gmail.com',
      url='https://django-ninja.rest-framework.com',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )



from ninja.responses import Response
from django.http import Http404
from django.urls import reverse
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    # if anyone konw a cleaner way to make mypy happy - welcome
    from ninja import NinjaAPI  # pragma: no cover


def home(request, api: "NinjaAPI"):
    "This view is mainly needed to determine the full path for API operations"
    docs_url = f"{request.path}{api.docs_url}".replace("//", "/")
    raise Http404(f"docs_url = {docs_url}")


def openapi_json(request, api: "NinjaAPI"):
    schema = api.get_openapi_schema()
    return Response(schema)


def swagger(request, api: "NinjaAPI"):
    return render(
        request,
        "ninja/swagger.html",
        {"api": api, "openapi_json_url": reverse(f"{api.urls_namespace}:openapi-json")},
    )


def render(request, template_name, context=None):
    """
    I do not relly want ninja to be required in INSTALLED_APPS for now
    that is why for now we use this render function to simulate django render
    """
    import os
    from django.template import Template, RequestContext
    from django.http import HttpResponse

    tpl_file = os.path.join(os.path.dirname(__file__), "../templates", template_name)
    with open(tpl_file) as f:
        tpl = Template(f.read())
    html = tpl.render(RequestContext(request, context))
    return HttpResponse(html)



from django.db import models


class Event(models.Model):
    title = models.CharField(max_length=100)
    start_date = models.DateField()
    end_date = models.DateField()

    def __str__(self):
        return self.title


class Client(models.Model):
    key = models.CharField(max_length=20, unique=True)



import pytest
from ninja import NinjaAPI
from django.test import Client


api = NinjaAPI()


@api.get("/test")
def method(request, query: str, count: int):
    return [query, count]


def test_schema(client: Client):
    assert client.get("/api/").status_code == 404
    assert client.get("/api/docs").status_code == 200
    assert client.get("/api/openapi.json").status_code == 200
    # TODO: more schema tests



import os
from ninja.openapi import get_schema
from typing import List, Optional, Tuple, Sequence, Union, Callable
from django.urls import reverse
from ninja.openapi.urls import get_openapi_urls
from ninja.router import Router
from ninja.errors import ConfigError
from ninja.constants import NOT_SET


class NinjaAPI:
    _registry: List[str] = []

    def __init__(
        self,
        *,
        title: str = "NinjaAPI",
        version: str = "1.0.0",
        description: str = "",
        openapi_url: Optional[str] = "/openapi.json",
        docs_url: Optional[str] = "/docs",
        urls_namespace: str = None,
        auth: Union[Sequence[Callable], Callable, object] = NOT_SET,
    ):
        self.title = title
        self.version = version
        self.description = description
        self.openapi_url = openapi_url
        self.docs_url = docs_url
        self.urls_namespace = urls_namespace or f"api-{self.version}"
        self.auth: Optional[Sequence[Callable]] = NOT_SET
        if auth is not None and auth is not NOT_SET:
            self.auth = isinstance(auth, Sequence) and auth or [auth]

        self._validate()

        self._routers: List[Tuple[str, Router]] = []
        self.default_router = Router()
        self.add_router("", self.default_router)

    def get(self, path: str, *, auth=NOT_SET):
        return self.default_router.get(path, auth=auth is NOT_SET and self.auth or auth)

    def post(self, path: str, *, auth=NOT_SET):
        return self.default_router.post(
            path, auth=auth is NOT_SET and self.auth or auth
        )

    def delete(self, path: str, *, auth=NOT_SET):
        return self.default_router.delete(
            path, auth=auth is NOT_SET and self.auth or auth
        )

    def patch(self, path: str, *, auth=NOT_SET):
        return self.default_router.patch(
            path, auth=auth is NOT_SET and self.auth or auth
        )

    def put(self, path: str, *, auth=NOT_SET):
        return self.default_router.put(path, auth=auth is NOT_SET and self.auth or auth)

    def api_operation(self, methods: List[str], path: str, *, auth=NOT_SET):
        return self.default_router.api_operation(
            methods, path, auth=auth is NOT_SET and self.auth or auth
        )

    def add_router(self, prefix, router):
        self._routers.append((prefix, router))

    @property
    def urls(self):
        return (
            self._get_urls(),
            "ninja",
            self.urls_namespace,
        )

    def _get_urls(self):
        result = get_openapi_urls(self)

        for prefix, router in self._routers:
            for path in router.urls_paths(prefix):
                result.append(path)
        return result

    @property
    def root_path(self):
        name = f"{self.urls_namespace}:api-root"
        return reverse(name)

    def get_openapi_schema(self, path_prefix=None):
        if path_prefix is None:
            path_prefix = self.root_path
        return get_schema(api=self, path_prefix=path_prefix)

    def _validate(self):
        skip_registry = os.environ.get("NINJA_SKIP_REGISTRY", False)
        if not skip_registry and self.urls_namespace in NinjaAPI._registry:
            msg = [
                "Looks like you created multiple NinjaAPIs",
                "To let ninja distinguish them you need to set either unique version or url_namespace",
                " - NinjaAPI(..., version='2.0.0')",
                " - NinjaAPI(..., urls_namespace='otherapi')",
                f"Already registered: {NinjaAPI._registry}",
            ]
            raise ConfigError("\n".join(msg))
        NinjaAPI._registry.append(self.urls_namespace)



from ninja.security.apikey import APIKeyQuery, APIKeyCookie, APIKeyHeader
from ninja.security.http import HttpBearer, HttpBasicAuth


def django_auth(request):
    if request.user.is_authenticated:
        return request.user
